package unsw891;

import unsw891.sml.*;
import unsw891.optimiser.*;
import unsw891.optimiser.global.*;
import unsw891.optimiser.local.*;
import unsw891.parser.*;

/**
 * The main class executed from command prompt
 */
public class SMLOptimiser {
        private static final double FUDGE_FACTOR = 0.05;

  	public static final boolean dontUseValidator = false;

	public static void main(String[] args) {
                int tlimit = Integer.parseInt(args[0]);
                long tcur, tgfinish, tstart = System.currentTimeMillis() / 1000;
                    

                System.err.println("SMLOptimiser started with time limit = "
                        + tlimit + ".");

		//Read input
		Document sourceDoc = Parser.parse(System.in);
	
	  	Document curBest = (Document)(sourceDoc.clone());	
		
		//white space optimiser (simple)
		curBest = OptimiserWrapper.optimise(new unsw891.optimiser.local2.LocalOptimiser(curBest), curBest);
		
		//white space optimiser (hard)
		curBest = OptimiserWrapper.optimise(new unsw891.optimiser.local8.LocalOptimiser(curBest), curBest);
	
	

                tcur = System.currentTimeMillis() / 1000;
                tgfinish = tstart + tlimit - 
                    Math.round((tcur - tstart)*(1 + FUDGE_FACTOR));

                System.err.println("SMLOptimiser - global optimiser has " +
                        (tgfinish - tstart) + " seconds to run.");
	
		//Run Document through global optimiser
		Document tmpDoc = (Document)(curBest.clone());
                
		try {

		  Optimiser opt = new GlobalOptimiser(tmpDoc);
		  Thread optThread = new Thread((Runnable)opt);
                  optThread.start();

                  while (tcur < tgfinish && !opt.isFinished()) {
                      try {
                          Thread.sleep(1000);
                      } catch (InterruptedException e) {
                          System.err.println("SMLOptimiser thread interrupted "
                              + "exception: " + e);
                      }

                      tcur = System.currentTimeMillis() / 1000;
                  }

                  if (!opt.isFinished())
                      System.err.println("GlobalOptimiser time limit has "
                              + "expired. Killed by SMLOptimiser thread!!!.");

                  tmpDoc = opt.outputResult();
                  optThread.stop();

		  if (dontUseValidator || Validator.SMLEquivalent(curBest, tmpDoc)) {
		  	System.err.println("SMLOPTIMISER: Accepted global optimisation!!");
			curBest = tmpDoc;
		  } else {
		  	System.err.println("Ran global optimiser, but rejected the optimisations");
			System.err.println("---output of rejected optimiser-------------");
			System.err.println(tmpDoc);
			System.err.println("--------------------------------------------");
		  }
		} catch (Exception e) {
		  e.printStackTrace();
		  System.err.println("Some exception was thrown when running global optimiser and validator");
		}
		
		//Run Document through local optimiser again
		//local optimisers
		curBest = OptimiserWrapper.optimise(new unsw891.optimiser.local.LocalOptimisers(curBest), curBest);
		
		
		//Print the output
		System.out.print(curBest.toString());
		
		System.err.println("Percentage Reduction:  " + (100.0*(sourceDoc.tagSize()-curBest.tagSize())/sourceDoc.tagSize()) + "%");
		
		System.exit(0);
	}
}
		
		
