package unsw891.optimiser.global;

import unsw891.sml.*;
import unsw891.optimiser.*;
import unsw891.util.*;
import java.util.Vector;

/**
 * Global optimiser, undertaken by Stephen
 */
public class GlobalOptimiser implements Optimiser, Runnable {

  private Document doc;
  private byte[] answer;
  private Shard[] shards;
  private Heuristics heurs;
  private boolean pleaseStop;
  private boolean finished;

  public GlobalOptimiser(Document doc) {
    pleaseStop = false;
    finished = false;
    this.doc = doc;
    answer = null;
  }

  public Document outputResult() {
  	if (answer == null) return doc;
	else return DnaCode.toDoc(answer,shards,new Heuristics(heurs));
  }

  public boolean isFinished() {
    return finished;
  }

  public void stop() {
    pleaseStop = true;
  }
  
  private void breakPoint() {
    while (pleaseStop) {
      try { this.wait(); }
      catch (InterruptedException e) {}
    }
  }
  
  public void resume() {
    pleaseStop = false;
    this.notifyAll();
  }
	
  public void run() {

    // first convert to a shard vector
    Vector shardsTemp = Converter.getShardVector(doc);
    
    // next fix up the whitespace
    //***************** DO THIS *************
    Vector shardsFixed = shardsTemp;
    
    // convert to an array & set up Heuristics
    int i;
    shards = new Shard[shardsFixed.size()];
    for (i = 0; i < shardsFixed.size(); i++) {
      shards[i] = (Shard) shardsFixed.elementAt(i);
    }
    
    heurs = new Heuristics(shards);
  
    answer = DnaCode.invent(shards,new Heuristics(heurs));
	System.err.println("Finished");
	finished = true;

  } // end of run method
  

} // end of class
