package unsw891.optimiser.global;

import unsw891.sml.*;
import unsw891.util.Debug;
import java.util.*;


// $Log: Heuristics.java,v $
// Revision 1.17  2001/07/29 10:50:08  lawry
// Okay, finished sortTag(), everything should work, hopefully it doesn't
// use that much more memory.
//
// So it now works for all tags except PL and U which it assumes to have
// a score of 0.
//
// Revision 1.16  2001/07/29 09:56:55  lawry
// Now Compiles, STUPID ME!!!
//
// Revision 1.14  2001/07/29 09:53:30  lawry
// Doc modification. Heuristics works with colour and size and stuff now.
//
// Last thing is to implement sortTag().
//
// Revision 1.13  2001/07/29 09:38:14  lawry
// Debugged and fixed many bugs in the heuristics engine, it should
// work properly now. I hope.
//
// Revision 1.9  2001/07/29 06:02:58  lawry
// Fixed some bugs and in process of implementing colour.
//
// Revision 1.8  2001/07/29 05:00:05  lawry
// Accepts an array instead of a vector now.
//
// Revision 1.7  2001/07/29 04:42:27  lawry
// HeurSize added a copy constructor.
//
// Heuristics now works in counting sizes. (there's still colour to go...)
// BUt needs massive, testing. I mean MASSIVE.
//
// Revision 1.6  2001/07/29 03:06:33  lawry
// Removed clone support. Use copy constructor instead. It's more efficient.
//
// Revision 1.5  2001/07/29 02:23:10  lawry
// The statistics part works for B EM I S and TT (PL is implied).
//
// Revision 1.4  2001/07/29 02:16:58  lawry
// Added copy constructor and clone() method as part of the interface.
// Copy constructor is more efficient than clone() method in use.
//


/**
 * stores information used by the heuristic function
 */
public class Heuristics {

    private Shard[] fDocShards;
    private int     fPosition;

    private int[]   fProperty;
    private Vector  fPropSize;
    private Vector  fPropColor;
    
    
    /** This value has to be larger than the largest value the tags
        are taking on. It's a hack, current at 30, will need to be
        changed if new tags added or changes are made to tag organisation.
        */
    public static final int MAX_TAGVAL = 30;


    /** Take a vector of shards, and from here calculate the
        information required. Only 1 pass is performed through the
        vector at the start, so this constructor may take a while
        to finish running.
        @param shards The vector of shards to gather info from. You
            should never change this vector once passed to this
            class.
        */
    public Heuristics(Shard[] shards) {
        fDocShards = shards;
        fPosition = 0;
        fPropSize = new Vector();
        fPropColor = new Vector();
        fProperty = new int[MAX_TAGVAL];
        
        performScan();
    }


    /** Copy constructor, use this to make an exact copy of this
        heuristic. This means the returned object will be a Heuristics
        class with update at the exact same position as this object.
        */
    public Heuristics(Heuristics h) {
        // shards vector is not copied. merely referenced.
        fDocShards = h.fDocShards;
        fPosition = h.fPosition;
        
        int i, vsize;
        
        // Size and colour property vector is deep copied.
        vsize = h.fPropSize.size();
        fPropSize = new Vector(vsize);
        for (i = 0; i < vsize; ++i)
            fPropSize.add(
                new HeurSize(
                    (HeurSize) h.fPropSize.get(i)
                )
            );

        vsize = h.fPropColor.size();
        fPropColor = new Vector(vsize);
        for (i = 0; i < vsize; ++i)
            fPropColor.add(
                new HeurColor(
                    (HeurColor) h.fPropColor.get(i)
                )
            );
        
        // Copy normal properties.
        vsize = h.fProperty.length;
        fProperty = new int[vsize];
        for (i = 0; i < vsize; ++i)
            fProperty[i] = h.fProperty[i];
    }





    /** Scans the vector to collect all statistics for later retrieval
        through methods.
        */
    protected void performScan() {
        int i, vsize = fDocShards.length;
        
        HeurSize hsz = null;
        HeurColor hcl = null;
        
        // Iterate through all shards in array.
        for (i = 0; i < vsize; ++i) {
            Shard s = fDocShards[i];

            if (s.isBold())
                ++fProperty[Document.TAG_B];
            if (s.isEm())
                ++fProperty[Document.TAG_EM];
            if (s.isItalic())
                ++fProperty[Document.TAG_I];
            if (s.isStrong())
                ++fProperty[Document.TAG_S];
            if (s.isTT())
                ++fProperty[Document.TAG_TT];
            
            /////////SIZE/////////////

            if (s.size > -1) {
                if (hsz == null) {         // create new obj.
                    hsz = new HeurSize(i);
                    ++hsz.sizeCount[s.size];
                } else {
                    ++hsz.sizeCount[s.size];
                }
            } else if (s.size == -1) {
                if (hsz != null) {      // end the obj
                    hsz.endpos = i - 1;
                    fPropSize.add(hsz);
                    hsz = null;
                }
            }

            /////////// COLOUR /////////////

            if (s.colour != 'x') {
                if (hcl == null) {      // create new heurcolor.
                    hcl = new HeurColor(i);
                    hcl.incr(s.colour);
                } else {
                    hcl.incr(s.colour);
                }
            } else if (s.colour == 'x') {
                if (hcl != null) {
                    hcl.endpos = i - 1;
                    fPropColor.add(hcl);
                    hcl = null;
                }
            }
        }
        // Since it is possible to have an entire document coloured or
        // set in size, we must end it here and add to vector.
        if (hsz != null) {
            hsz.endpos = vsize - 1;
            fPropSize.add(hsz);
        }
        
        if (hcl != null) {
            hcl.endpos = vsize - 1;
            fPropColor.add(hcl);
        }
    }
    
    
    
  
    /** Notify that we've just finished dealing with this Shard.
        Advancement starts from index 0 (the very first shard in the
        vector) and moves onwards until shard number size()-1.
        
        <p>Actually, you don't need the parameter as it is not used,
        position (and therefore shard is kept internally. If you do
        supply a shard, it will be checked for equality, to make sure
        we are in the right position.
        */
    public void update(Shard sh) {
        Shard s = fDocShards[fPosition];
        if (sh != null)
            Debug.assert(sh == s || sh.equals(s),
                "Heuristics::update() param " +
                "is different from what I think internally.");
        
        if (s.isBold())
            --fProperty[Document.TAG_B];
        if (s.isEm())
            --fProperty[Document.TAG_EM];
        if (s.isItalic())
            --fProperty[Document.TAG_I];
        if (s.isStrong())
            --fProperty[Document.TAG_S];
        if (s.isTT())
            --fProperty[Document.TAG_TT];

        ///////////////// SIZES //////////////////
        
        if (s.size > -1) {
            // Note that heursize vector is already sorted by position.
            HeurSize hsz;
            if (fPropSize.size() > 0) {
                hsz = (HeurSize) fPropSize.firstElement();

                while (fPosition > hsz.endpos) {
                    // this element is now useless, get rid of it.
                    fPropSize.removeElementAt(0);

                    if (fPropSize.size() > 0)
                        hsz = (HeurSize) fPropSize.firstElement();
                    else
                        break;
                }

                // Found a HeurSize obj to work with.
                if (fPropSize.size() > 0 && fPosition >= hsz.startpos)
                    --hsz.sizeCount[s.size];
            }
        }
        
        ///////////////////// COLOUR ///////////////////////
        
        if (s.colour != 'x') {
            // Note that heurcolor vector is already sorted by position.
            HeurColor hcl;
            if (fPropColor.size() > 0) {
                hcl = (HeurColor) fPropColor.firstElement();

                while (fPosition > hcl.endpos) {
                    // this element is now useless, get rid of it.
                    fPropColor.removeElementAt(0);

                    if (fPropColor.size() > 0)
                        hcl = (HeurColor) fPropColor.firstElement();
                    else
                        break;
                }

                // Found a HeurColor obj to work with.
                if (fPropColor.size() > 0 && fPosition >= hcl.startpos)
                    hcl.decr(s.colour);
            }
        }
        
        ++fPosition;
    }
  

    /** Returns a property count for the B, EM, I, S and TT tags. The
        number returned is how many shards have the given tag (tagType)
        in the <strong>entire</strong> shard array passed to the
        constructor.
        <p>Underfined result is any other tagType is used other than
        the 5 mentioned above.
        */
    public int getPropertyCount(int tagType) {
        return fProperty[tagType];
    }
    
    
    
    /** Return a property count for the size tags, ie TAG_0 to TAG_9.
        The returned value is how many continuous shards have the given
        size (use TAG_4 etc to specify size in param) from the current
        position, in a continuous band of (any) sized shards.
        */
    public int getSizeCount(int sizeTag) {
        int size = sizeTag - Document.TAG_0;
        Debug.assert(size >= 0 && size < HeurSize.SIZE_COUNT,
            "Invalid tag passed to Heuristics::getSizeCount()");
        
        // Note that heursize vector is already sorted by position.
        HeurSize hsz;
        if (fPropSize.size() > 0) {
            hsz = (HeurSize) fPropSize.firstElement();
        
            while (fPosition > hsz.endpos) {
                // this element is now useless, get rid of it.
                fPropSize.removeElementAt(0);

                if (fPropSize.size() > 0)
                    hsz = (HeurSize) fPropSize.firstElement();
                else
                    return 0; // since we don't have any more to work with.
            }
            
            // Found a HeurSize obj to work with.
            if (fPosition >= hsz.startpos)
                return hsz.sizeCount[size];
            else
                return 0; // we are not in range yet.
        
        } else {
            // since we don't have anymore info to work with, sizes will be
            // zero from now on.
            return 0;
        }
    }
    
    
    
    /** Same as the other getColourCount(), but this one accepts the
        colour tags (ie TAG_w, TAG_g etc).
        */
    public int getColorCount(int colourTag) {
        switch (colourTag) {
        
            case Document.TAG_r:
                return getColorCount('r');
            
            case Document.TAG_g:
                return getColorCount('g');
            
            case Document.TAG_b:
                return getColorCount('b');
            
            case Document.TAG_c:
                return getColorCount('c');
            
            case Document.TAG_m:
                return getColorCount('m');
            
            case Document.TAG_y:
                return getColorCount('y');
            
            case Document.TAG_k:
                return getColorCount('k');
            
            case Document.TAG_w:
                return getColorCount('w');
            
            default:
                Debug.assert(false, "HeurColor: bad colour to count");
                return 0;
        }
    }
    
    

    /** Return a property count for the colour tags, ie TAG_r and TAG_w
        etc. The returned value is how many shards have the given
        colour in a continuous band of (any) colour shards. Use
        characters to specify colour in param, eg 'c' for cyan, 'k' for
        black etc. The count starts from the current position.
        */
    public int getColorCount(char colour) {
        // Note that heurcolor vector is already sorted by position.
        HeurColor hcl;

        if (fPropColor.size() > 0) {
            hcl = (HeurColor) fPropColor.firstElement();
        
            while (fPosition > hcl.endpos) {
                // this element is now useless, get rid of it.
                fPropColor.removeElementAt(0);

                if (fPropColor.size() > 0)
                    hcl = (HeurColor) fPropColor.firstElement();
                else
                    return 0; // since we don't have any more to work with.
            }
            
            // Found a HeurColor obj to work with.
            if (fPosition >= hcl.startpos)
                return hcl.getVal(colour);
            else
                return 0; // we are not in range yet.
        
        } else {
            // since we don't have anymore info to work with, sizes will be
            // zero from now on.
            return 0;
        }
    }
    
  
  
    /** <em>
        * Here is the deal. We realise that we have to open a few tags.
        * For example, the next text must be bold, and emphasised. But,
        * should we open the bold tag first, or the emphasised tag first?
        * Probably, the tag which occurs more frequently in the future 
        * should be opened first. This function is given a Vector of
        * Document objects, each of which is a single tag with no children.
        * The function should return a Vector where the tags have been
        * sorted so that the most frequently occuring tag is at the
        * beginning.

        </em><p>
        Okay, enough contemplating :-) this method will get the properties
        of the tags passed through the vector, and sort them so that the
        most frequently appearing tag will be first (using the heuristic
        property returned by other method in this object).
        
        <p>Due to difficulty in assigning scores for the U and PL tag,
        they are ignored during the sorting (ie assigned a score of 0).
        So best you don't include these 2 tags.
        */   /*
          notes to self:
             // might need to do some jiggery-pokery so that we handle
             // <U> tags and <EM> tags properly.

             // also, try to take account of different length tags?

             // take into account undefined sizes and colours, reducing 
             // the property count.
             */
    public Vector sortTags(Vector tags) {
        int i, vsize = tags.size();
        
        Vector keep = new Vector(vsize);
        
        // Create a vector with the scores obj
        for (i = 0; i < vsize; ++i) {
            int docTag = ((Document) tags.get(i)).getType();
            int tagScore;
            
            if (Document.TAG_0 <= docTag && docTag <= Document.TAG_9) {
                // THIS IS SIZE
                tagScore = getSizeCount(docTag);

            } else if (Document.TAG_r <= docTag && docTag <= Document.TAG_w) {
                // THIS IS COLOUR
                tagScore = getColorCount(docTag);
            
            } else if (docTag == Document.TAG_PL || docTag == Document.TAG_U) {
                tagScore = 0;
            
            } else {
                tagScore = getPropertyCount(docTag);
            }
            
            keep.add(new HeurScore(docTag, tagScore));
        }
        Collections.sort(keep); //NOTE THIS IS IN ASCENDING.
        Vector retVal = new Vector(vsize);
        
        for (i = vsize -1; i >= 0; --i) {
            HeurScore hs = (HeurScore) keep.get(i);
            retVal.add(new Document(hs.tag));
        }
        
        return retVal;
    }
  
  
}
