package unsw891.optimiser.local1;

import unsw891.sml.*;
import unsw891.optimiser.*;
import java.util.*;
import unsw891.util.*;

/**
 * Performs EM elimination. Requires combined use with white space optimiser and redundency optimiser
 */
public class LocalOptimiser implements Optimiser {

	private Document doc;
	
	public LocalOptimiser(Document doc) {
		this.doc = doc;
	}

	/**
	 * <MODIFICATION LOG>
	 * Friday 5:02pm - John Lai - created but not yet complete. Need to add some
	 * local optimisations
	 */
	public Document outputResult() {
	
		//System.err.println("unsw891.local1.LocalOptimiser.outputResult() not yet implemented");
		
		//Run your optimisation
		doc.genLinks();
		optimiseFromRoot(doc);
		return doc;
	}
	
	/**
	 *Eliminate unnecesary EMs. pass each child to optimise function
	 */
	private void optimiseFromRoot(Document doc) {
	  for (int i=0; i<doc.children.size(); i++) {
	  	int oldSize = doc.children.size();
		optimise( (Document)(doc.children.get(i)) );
		  
		//check if this doc.children vector had changed
		if (oldSize != doc.children.size()) {
		  i--;
		}
	  }
	}
	
	/**
	 *Eliminates unnecessary EMs.
	 */
	private void optimise(Document doc) {
	  //doc shouldn't be root
	  Debug.assert(doc.type!=Document.TAG_ROOT, "local1.LocalOptimiser.optimise() encountered a root document");
	  
	  if (doc.children.size()==0) return;
	  
	  if (doc.type!=Document.TAG_EM) {
	  	for (int i=0; i<doc.children.size(); i++) {
		  int oldSize = doc.children.size();
		  optimise( (Document)(doc.children.get(i)) );
		  
		  //check if this doc.children vector had changed
		  if (oldSize != doc.children.size()) {
		  	i--;
		  }
		}
		return;
	  }
	  
	  //this doc is an EM and has at least one child
	  
	  //check if the first child is an EM
		
	  Document firstChild = (Document)(doc.children.get(0));
	  if (firstChild.type==Document.TAG_EM) {
	  	int index = doc.parent.children.indexOf(doc);
		
		doc.children.remove(0);
		
		firstChild.parent = doc.parent;
		doc.parent.children.addAll(index, firstChild.children);
		
		
	  } else {
	  	Document lastChild = (Document)(doc.children.lastElement());
		if (lastChild.type==Document.TAG_EM) {
	  	  int index = doc.parent.children.indexOf(doc);
		
		  doc.children.remove(doc.children.size()-1);
		  	
		  lastChild.parent = doc.parent;
		  doc.parent.children.addAll(index, lastChild.children);
		} else {
		  for (int i=0; i<doc.children.size(); i++) {
		  	int oldSize = doc.children.size();
  		  	optimise( (Document)(doc.children.get(i)) );
		  
		  	//check if this doc.children vector had changed
		  	if (oldSize != doc.children.size()) {
		  	  i--;
		  	}
		  }
		}
	  }
	}
	  	
	
	
	public boolean isFinished() {
		System.err.println("unsw891.local1.LocalOptimiser.isFinished() not yet implemented");
		return true;
	}
	
	public void run() {
		System.err.println("unsw891.local1.LocalOptimiser.run() not yet implemented");
		
	}

	public void stop() {
		System.err.println("unsw891.local1.LocalOptimiser.stop() not yet implemented");
		
	}
	
	public String toString() {
	  return "John Lai - EM elimination";
	}
}
