package unsw891.optimiser.local8;

import unsw891.sml.*;
import unsw891.optimiser.*;
import java.util.Vector;

/**
 * Jason's Local Optimiser
 *
 * Contents:
 * 1 White Space Compressor
 * 1 White Space Simplifier
 *
 * Version 0.3 (2/3 written) Created: 27/7/01
 * Last Modified: 28/7/01
 *
 * My Work Log
 * 28/7/01 - fixed up a little if statement that made everything wrong
 * 28/7/01 - completed the base optimisation methods and other code
 * 28/7/01 - migrated base code into this file
*/

public class LocalOptimiser implements Optimiser {

	private Document doc;
	private Document currentDoc;
        	
	public LocalOptimiser(Document doc) {
		this.doc = doc;
		currentDoc = doc;
                
	}

	/**
	 * <MODIFICATION LOG>
	 * Sunday 7:15   - Sherman Lo - Fixed bug with spaces not being counted in prevCon in tt
         * Saturday 8:10 - Jason Au - fixed up tt mode
	 * Saturday 1:59 - Jason Au - made some minor corrections
	 * Saturday 1:14 - Jason Au - implemented outputResult()
	 * Saturday 1:12 - jason Au - finished coding the base optimisation code
	 * Saturday 10:30 - Jason Au - moved code from my home file into this file
	 * Friday 5:02pm - John Lai - created but not yet complete. Need to add some
	 * local optimisations
	 */
	public Document outputResult() {
	
		System.err.println("unsw891.local2.LocalOptimiser.outputResult() is now implemented");
		
		//Run your optimisation
		
		descendTree(new SpaceContext(),null);
		
		return doc;
	}
		
	
	public boolean isFinished() {
		System.err.println("unsw891.local2.LocalOptimiser.isFinished() not yet implemented");
		return true;
	}
	
	public void run() {
		System.err.println("unsw891.local2.LocalOptimiser.run() not yet implemented");
		
	}

	public void stop() {
		System.err.println("unsw891.local2.LocalOptimiser.stop() not yet implemented");
		
	}
	
	/**
	 * Jason
	 * Tree descender - goes through each child sequentially and does
	 * the optimisations on each one.
	 */
	private SpaceContext descendTree(SpaceContext currentCon, SpaceContext prevCon) {
	
	   // save current doc
	   Document here = currentDoc;
	   
	   // make a new children vector
	   Vector newChildren = new Vector();
	   
	   // make this
	   SpaceContext childCon = prevCon;
	   
	   // go through the children
	   for(int i=0; i < here.children.size() ; i++) {
	   
	      // change working set to next child
	      currentDoc = (Document) here.children.elementAt(i);
	      
	      // optimise it, and get the child's context
	      childCon = optimiseChild(currentCon,childCon);
	      
	      // prune any empty leaves
	      if(currentDoc.getType() == Document.TAG_TEXT) {
                 if(currentDoc.getText().equals("")) {
                    continue;
                 }
              }
	      
	      // otherwise add to new vector
	      newChildren.addElement(currentDoc);
           }
	   // reallocate everything
           here.children = newChildren;
	   currentDoc = here;
	   return childCon;
	   
	}
	
	/**
	 * Jason
	 * Optimiser
	 * if it isn't a leaf, it records the context and engages the descender
	 * if it is a leaf, compresses any whitespace, and then checks
	 * whether simplifications can be made
	 */
	private SpaceContext optimiseChild(SpaceContext currentCon, SpaceContext prevCon) {
	
	   int type = currentDoc.getType();
	   if(type == Document.TAG_TEXT) {
	      
	      // forgot this little if statement.  oops!
              if(!currentCon.tt) {
                 // compression code
	         boolean wsOn = false;
	         String text = currentDoc.getText();
	         String outStr = "";
                 for(int j=0; j<text.length();j++) {
                    char ch = text.charAt(j);
                    if(Character.isWhitespace(ch)|| ch == '\t') {
                       if(wsOn) {
                          continue;
                       } else {
                          ch = ' ';
                          wsOn = true;
                       }
                    } else {
                       wsOn = false;
                    }
                    outStr = outStr.concat((new Character(ch)).toString());
                 }
              
	      // simplification code
	      
	      // do simplify if and only if
	      // 1) there is a previous context to compare with
	      // 2) the current text begins with whitespace
	      // 3) the enable flag is on
	      // 4) the space contexts are compatible
	         if(prevCon != null &&
	            outStr.startsWith(" ") &&
   		    prevCon.simplify_enable &&
                    currentCon.isCompatible(prevCon) ) {
		 
		 // because the string is already compressed
		 // we can do this with a certain amount of safety
		    outStr = outStr.substring(1);
		 
	         }


             currentDoc.text = outStr;
              }
	         currentCon.simplify_enable = currentDoc.text.endsWith(" ") ||
                                          currentDoc.text.endsWith("\n") ||
                                          currentDoc.text.endsWith("\r") ||
                                          currentDoc.text.endsWith("\t");
              return currentCon;
	   } else if(type == Document.TAG_TT) {
	      currentCon.tt = true;
	      SpaceContext tempCon = descendTree(currentCon,prevCon);
              currentCon.tt = false;
              return tempCon;
	   } else if(type == Document.TAG_U) {
	      currentCon.incU();
	      SpaceContext tempCon = descendTree(currentCon,prevCon);
              currentCon.decU();
              return tempCon;
	      
	   } else if(type >= Document.TAG_r && type <= Document.TAG_w) {
               int oldCol = currentCon.colour;
	      currentCon.setColour(type);
	      SpaceContext tempCon = descendTree(currentCon,prevCon);
	      currentCon.colour=oldCol;
              return tempCon;

	   } else if(type >= Document.TAG_0 || type <= Document.TAG_9){
               int oldsize = currentCon.size;
	      currentCon.setSize(type);
	      SpaceContext tempCon = descendTree(currentCon,prevCon);
	      currentCon.setSize(oldsize);
              return tempCon;
	      
	   } else if(type == Document.TAG_PL) {
               boolean oldtt = currentCon.tt;
               int oldU = currentCon.u;
               int oldsize = currentCon.size;
	      currentCon.tt = false;
	      currentCon.u = 0;
	      currentCon.setSize(Document.TAG_0);
	      SpaceContext tempCon = descendTree(currentCon,prevCon);
              currentCon.tt = false;
	      currentCon.u = oldU;
	      currentCon.setSize(oldsize);
              return tempCon;
	   }
	   return descendTree(currentCon,prevCon);
	   
	}
	
	public String toString() {
	   return "White Space Optimiser Hard: Written by Jason Au for ICFP2001";
	}
}

