package unsw891.optimiser.local8;

import unsw891.sml.*;

/**
 * Jason's SpaceContext, a handy container object
 * for boolean values which can be used to propagate
 * between children
 *
 */
class SpaceContext {

   /*
    * Because only I am using this class, I can afford to be
    * lazy with protection.
    */
   boolean tt;
   int u;
   int size;
   int colour;
   boolean simplify_enable;
   
   /**
    * Constructor
    */
   public SpaceContext() {
      tt = false;
      u = 0;
      size = 0;
      colour = Document.TAG_w;
      simplify_enable = false;
   }
   
   /*
    * Sets the colour
    * @param tag the tag denoting the colour context
    */
   public void setColour(int tag) {
      
      if (tag >= Document.TAG_r && tag <= Document.TAG_w) {
         colour = tag;
      } else {
         return;
      }
   }
   
   /*
    * Sets the size
    * @param tag the tag denoting the size context
    */
   public void setSize(int tag) {
      if (tag >= Document.TAG_0 && tag <= Document.TAG_9) {
         size = tag;
      } else {
         return;
      }
   }
   
   /*
    * Tests whether the current Space Context and another Space Context
    * are compatible for white space simplification
    * @param other the other Space Context
    * @return the test result as a boolean
    */
   public boolean isCompatible(SpaceContext other) {
   
      // two space contexts are compatible if and only if their
      // space context properties match.
      
      if(other == null) {
         return false;
      }
      
      // colour only matters if u is non-zero
      boolean colourOn = (this.u == 0 &&
                          other.u == 0) ||
                         (this.u == other.u &&
                          this.colour == other.colour);
      
      return (this.tt == other.tt &&
	      this.size == other.size &&
	      colourOn);
   }
   
   /*
    * Increment U
    */
   public void incU() {
      u = Math.min(3,(u+1));
   }
   
   /*
    * Decrement U
    */
   public void decU() {
      u = Math.max(0,(u-1));
   }
   
}
