package unsw891.sml;

/**
 * Represents a context of a SML evaluating machine.
 * Used in global optimizers and validators.
 * <MODIFICATION LOG>
 * Sunday 12noon - John - fixed equals() method - didn't compare s field
 * 20:23, Friday - Jordan - Added equals() and whiteSpaceEquals();
 */
public class Context {
	
	/**
	 *Since "Two documents are equivalent if they have the same meaning
	 *in every possible root context, (i.e. for all values of size and
	 *color)", the following two constants have been defined to distinguish
	 *the size and colour of the root context. They're essentially like
	 *an INVALID value.
	 */
	public static final int ROOT_CONTEXT_SIZE = -1;
	public static final int ROOT_CONTEXT_COLOUR = 'x';
	
	/**
	 * Bold
	 */
	public boolean b;
	
	/**
	 * Emphasis
	 */
	public boolean em;
	
	/**
	 * Italics
	 */
	public boolean i;
	
	/**
	 * Strong
	 */
	public boolean s;
	
	/**
	 * teletype
	 */
	public boolean tt;
	
	/**
	 * underline
	 */
	public int u;
	
	/**
	 * text size
	 */
	public int size;
	
	/**
	 * text colour
	 */
	public char colour;
	
	
	/**
	 * Constructor - creates a root context
	 */
	public Context() {
		b  = false;
		em = false;
		i  = false;
		s  = false;
		tt = false;
		u  = 0;
		
		size = ROOT_CONTEXT_SIZE;
		colour = ROOT_CONTEXT_COLOUR;
	}
		
	public Object clone() {
		Context c = new Context();
		
		c.b      = b;
		c.em     = em;
		c.i      = i;
		c.s      = s;
		c.tt     = tt;
		c.u      = u;
		c.size   = size;
		c.colour = colour;
		
		return c;
	}


  	/**
	 *<MODIFICATION LOG>
  	 *Sunday 12noon - John - fixed bug. Didn't compare the s field
	 */
	public boolean equals(Context c) {
		return ( c.b  == b  &&
			 c.em == em &&
			 c.i  == i  &&
			 c.s  == s  &&
			 c.tt == tt &&
			 c.u  == u  &&
			 c.size == size &&
			 c.colour == colour );
	}

}

