package unsw891.tester;

import unsw891.parser.*;
import unsw891.optimiser.*;
import unsw891.optimiser.local.*;
import unsw891.optimiser.global.*;
import unsw891.sml.*;
import java.io.*;
import java.util.*;

/*
 *Usage: java OptimiserTester [tests] inputFileNames...
 *
 *Tests currently consists of any combination of:
 *  *a - local1
 *  *b - local2
 *  *c - local3
 *  *d - local4
 *  *g - local7
 *  *z - global tester
 *  *0 - buggy optimiser for testing validator (replaces first non <I> with <I>
 *
 * The test input files will get run in order through the optimisers:
 * ie. java OptimiserTestr bd test1 test2
 * will run local2 and local4 in that order on the text in the file test1, then run them on the file test2.
 */

public abstract class OptimiserTester {

    OptimiserTester tester = null;
    static OptimiserTester startTest;
    Optimiser optimiser;
    static Vector testFiles = new Vector();

    OptimiserTester() {
	
    }
    
    OptimiserTester(OptimiserTester tester) {
	this.tester = tester;
    }

    public static void main(String[] args) {
	if (args.length > 0) {
	    String tests = args[0];
	    makeTests(tests);
	    System.err.println("Tests made");
	    if (args.length == 1) {
		//Maybe check for .tst files and add them to queue??
	    } else {
		for (int i = 1; i < args.length; i++) {
		    Document curDoc = parse(args[i]);
		    startTest.runTest(curDoc);
		}
	    }
	} else {
	    System.err.println("Incorrect usage");
	    System.exit(1);
	}
    }

    private static void makeTests(String tests) {
	OptimiserTester curTester = null;
	OptimiserTester prevTester = null;
	for (int i = 0; i < tests.length(); i++) {
	    char curChar = tests.charAt(i);
	    switch(curChar) {
	    case 'a':
		if (prevTester == null) {
		    curTester = new Tester1();
		} else {
		    curTester = new Tester1(prevTester);
		}
		prevTester = curTester;
		break;
	    case 'b':
		if (prevTester == null) {
		    curTester = new Tester2();
		} else {
		    curTester = new Tester2(prevTester);
		}
		prevTester = curTester;
		break;
	    case 'c':
		if (prevTester == null) {
		    curTester = new Tester3();
		} else {
		    curTester = new Tester3(prevTester);
		}
		prevTester = curTester;
		break;
	    case 'd':
		if (prevTester == null) {
		    curTester = new Tester4();
		} else {
		    curTester = new Tester4(prevTester);
		}
		prevTester = curTester;
		break;
	    case 'e':
		
	    case 'f':
		

		break;
	    case 'g':
		/*		if (prevTester == null) {
		    curTester = new Tester7();
		} else {
		    curTester = new Tester7(prevTester);
		}
		prevTester = curTester;
		break;*/
	    case 'z':
		if (prevTester == null) {
		    curTester = new GlobalTester();
		} else {
		    curTester = new GlobalTester(prevTester);
		}
		prevTester = curTester;
		break;
	    case '0':
		/*if (prevTester == null) {
		    curTester = new Tester5();
		} else {
		    curTester = new Tester5(prevTester);
		}
		prevTester = curTester;
		break;*/
	    default:
		System.err.println("Bleh");
	    }
	}
	startTest = prevTester;
    }

    public Document runTest(Document curDoc) {
	if (tester != null) {
	    curDoc = tester.runTest(curDoc);
	}
	if (curDoc != null) {
	    setupTest(curDoc);
	    return test(curDoc);
	} else {
	    System.err.println("Bleh");
	    return null;
	}
    }

    private static Document parse(String fileName) {
	try {
	    return Parser.parse(new FileInputStream(fileName));
	}
	catch (FileNotFoundException e) {
	    System.err.println(e);
	    return null;
	}
    }

    protected abstract void setupTest(Document doc);

    protected Document test(Document doc) {
	System.err.println("===============================");
	System.err.println("INPUT:");
	System.err.println(doc);
	Document startDoc = (Document) doc.clone();
	int startSize = doc.size();
	long startTime = System.currentTimeMillis();
	optimiser.run();
	Document result = optimiser.outputResult();
	System.err.println("OUTPUT:");
	System.out.print(result);
	System.err.println();
	long endTime = System.currentTimeMillis();
	System.err.println("Testing: " + optimiser);
	System.err.println("Input Size: " + startSize);
	System.err.println("Output Size: " + result.size());
	System.err.println("Percentage Reduction: " + (((double) startSize -  result.size()) / (double) startSize) * 100);
	System.err.println("CPU Time Taken: " + (endTime - startTime));
	System.err.println("Correctness Testing: " + Validator.SMLEquivalent(startDoc, result));
	System.err.println("===============================");
	return result;
    }

}





