package unsw891.optimiser.global;

import unsw891.sml.*;
import java.util.*;


// This is a stub file


/**
 * unravelling the meaning of the DNA code
 */
public class DNA {


  ShardStack stack;
  Shard nextShard;
  Heuristics heurs;
  int[] costs;
  int[] codesSorted;

  /**
   * This constructor gives the DNA object all the info
   * it needs to do it's funky thang!
   */
  public DNA(ShardStack stack, Shard nextShard, Heuristics heurs) {
  	
    // it is not envisaged that Heuristics will be used, but they
    // are included just in case
	this.stack = stack;
	this.nextShard = nextShard;
	this.heurs = heurs;
	this.costs = null;
	this.codesSorted = null;
  }
  
  
  /**
   * @return the cost of each code. Higher numbers mean worse codes.
   */
  public int[] getCosts() {
  	if (costs == null) calcCosts();
    return costs;
  }
  
  void calcCosts() {
  
  	// okay, the task is to check what happens depending on how many things
	// we pop off the stack.
	
	// convert the stack to a vector
	Vector stackVec = stack.toVector();
	Shard shardPL = new Shard(stack.peek());
	shardPL.flags = 0; // applying the plain tag
	
	// okay, now we can do the comparisons and work out which is better
	costs = new int[stackVec.size() + 1];
	for (int i = 0; i < stackVec.size(); i++) {
	  costs[i+1] = ((Shard)stackVec.elementAt(i)).changeCost(nextShard);
	}
	costs[0] = shardPL.changeCost(nextShard) + 9;
	
	// but there are some extra issues for plain tags?
	// we'll ignore them for now, until I remember what they were.
  
  }
    
    
  /**
   * @return the codes in sorted order. Codes occuring first are better.
   */
  public int[] getSortedCodes() {
	if (codesSorted == null) sortCodes();
    return codesSorted;
  }
  
  void sortCodes() {
  	if (costs == null) calcCosts();
  	long[] toSort = new long[costs.length];
	for (int i = 0; i < costs.length; i++) {
	  toSort[i] = ((long) costs[i] << 32) | i;
//	  System.err.println("cost, code: "+costs[i]+", "+i);
	}
	java.util.Arrays.sort(toSort);
  	
	codesSorted = new int[costs.length];
	for (int i = 0; i < costs.length; i++) {
	  codesSorted[i] = (int) (toSort[i] & 0xffffffff);
	}
  }
  
  
  /**
   * @return Given a dnaCode, this returns a Vector of tags needing
   * to be opened, and also pops the Stack appropriately.
   */
  public Vector getNewTags(int dnaCode) {
  
//  	System.err.println("Getting new tags for dna code: "+dnaCode);
  
    // pop off the stack until you've gone back far enough,
    // then add.
    int comparison;
    Document newNode;
	Shard currentS = stack.peek();
    Vector newTags = new Vector();
	
  	if (dnaCode == 0) {
	  newNode = new Document(Document.TAG_PL);
	  currentS = new Shard(currentS);
	  currentS.flags = 0;
	  stack.push(currentS);
	  newTags.add(newNode);
	}
  	for (int i = 0; i < dnaCode - 1; i++) {
      stack.pop();
      currentS = stack.peek();
    }

  	// we'll loop here if the code was wrong
  	while ((comparison = currentS.compare(nextShard)) == -1) {
      // go back one tag
      stack.pop();
      currentS = stack.peek();
    }


    // Work out which tags to add, by examining the
    // output of the Shard.compare method.

    // we store the tags to be added in a vector, as documents
    // with no children

    if ((comparison & 0x200) > 0) {  // size
      newTags.add(new Document(nextShard.size + Document.TAG_0));
    }
    if ((comparison & 0x100) > 0) {  // colour
      newTags.add(new Document(nextShard.colour));
    }
    if ((comparison & Shard.BOLD) > 0) {
      newTags.add(new Document(Document.TAG_B));
    }
    if ((comparison & Shard.ITALIC) > 0) {
      newTags.add(new Document(Document.TAG_I));
    }
    if ((comparison & Shard.TT) > 0) {
      newTags.add(new Document(Document.TAG_TT));
    }
    if ((comparison & Shard.STRONG) > 0) {
      newTags.add(new Document(Document.TAG_S));
    }
    if ((comparison & Shard.EM) > 0) {
      newTags.add(new Document(Document.TAG_EM));
    }
    for (int j = 0; j < (comparison & Shard.UNDER); j++) {
      newTags.add(new Document(Document.TAG_U));
    }
  
    return newTags;
  }
  
  


}
